/*
 * sat68kdb: Sega Saturn MC68EC000 Debugger
 * Bart Trzynadlowski
 */

/*
 * sat68kdb.c: Main module
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "satintrf.h"
#include "debug.h"

#define VERSION "0.0"


/*
 * Utility Functions
 */

void Error(char *input, ...)
{
    va_list arg;
    char    text[256];

    va_start(arg, input);
    vsprintf(text, input, arg);
    va_end(arg);
    fprintf(stderr, "68kdb: %s", text);
    exit(1);
}

void Warning(char *input, ...)
{
    va_list arg;
    char    text[256];

    va_start(arg, input);
    vsprintf(text, input, arg);
    va_end(arg);
    fprintf(stderr, "68kdb: Warning: %s", text);
}

int FindV(char *option, int p, int h, int u, int argc, char **argv, int m)
{
    static INT32    t[128] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
                              };
    INT32           i;
    char            *c;
          
    if (m)  /* mark as touched */
    {
        if (m < 128)
            t[m] = 1;
        return 0;
    }
    if (argc > 128)
        argc = 128;     /* maximum this function can handle is 128 */

    if (u)  /* find first untouched element */
    {
        for (i = 1; i < argc; i++)
        {
            if (!t[i])      /* 0 indicates untouched */
                return i;
        }
        return 0;
    }

    if (p)  /* find option and return integer value following it */
    {
        for (i = 1; i < argc; i++)
        {        
            if (strcmp(argv[i], option) == 0)       /* found */
            {
                if (i >= (argc - 1))        /* bounds! */
                    return 0;
                t[i + 1] = t[i] = 1;        /* touched */
                if (!h)
                    return atoi(argv[i + 1]);
                else
                    return strtoul(argv[i + 1], &c, 16);
            }
        }
        return 0;                       /* no match */
    }
    else    /* find option and return position  */    
    {
        for (i = 1; i < argc; i++)
        {        
            if (strcmp(argv[i], option) == 0)
            {
                t[i] = 1;
                return i;       /* found! return position */
            }
        }                              
        return 0;
    }
    return 0;
}


/*
 * main()
 */

int main(int argc, char **argv)
{
    FILE    *fp;
    UINT8   *mem;
    UINT32  i, j, port;

    if (argc <= 1 || FindV("-?", 0, 0, 0, argc, argv, 0) || FindV("-h", 0, 0, 0, argc, argv, 0))
    { 
        puts("sat68kdb Version "VERSION" by Bart Trzynadlowski: Sega Saturn MC68EC000 Debugger");
        puts("Usage:      sat68kdb <file> [options]");
        puts("Options:    -?,-h       Show this help text");
        puts("            -p <port>   Base port of PC Comm Card [Default=300]");
        exit(0);
    }

    port = FindV("-p", 1, 1, 0, argc, argv, 0);

    i = FindV(NULL, 0, 0, 1, argc, argv, 0);
    if (!i) Error("No input file specified\n");
    else
    {
        if ((fp = fopen(argv[i], "rb")) == NULL)
            Error("Failed to open input file: %s\n", argv[i]);
    }

    fseek(fp, 0, SEEK_END);
    j = ftell(fp);
    if (j < 2)      /* must be at least 2 bytes */
        Error("File is too small (%d bytes), 2 bytes is the minimum\n", j);
    if (j & 1)      /* must be even */
    {
        Warning("File size is odd (%d bytes), padding out to %d bytes\n", j, j + 1);
        j++;
    }
    rewind(fp);

    if ((mem = malloc(j)) == NULL)
        Error("Out of memory!\n");

    fread(mem, sizeof(UINT8), j, fp);
    fclose(fp);

    /*
     * Initialize Saturn
     */
    
    puts("sat68kdb Version "VERSION"");
    printf("Establishing link with Sega Saturn... ");
    sat68kdb_Init(port, mem, j);
    puts("OK\n");    
    puts("Type ? for usage help, q to quit");

    DebugSat68K(NULL, NULL);

    /*
     * Deinitialize Saturn
     */

    sat68kdb_Deinit();

    /*
     * Print data statistics
     */

    printf("\nTotal bytes transferred: %d\n", sat68kdb_bytes_transferred);

    free(mem);
    return 0;
}
